//----------------------------------------------------------------------
//
// This program test the behavior of pset::Registry.
//
// $Id: Registry_t.cpp,v 1.12 2008/07/01 18:22:17 rpw Exp $
//----------------------------------------------------------------------
#include <cassert>
#include <cmath>
#include <iostream>
#include <vector>

#include "boost/thread.hpp"

#include "DataFormats/Provenance/interface/ParameterSetID.h"
#include "FWCore/Utilities/interface/EDMException.h"
#include "FWCore/ParameterSet/interface/ParameterSet.h"
#include "FWCore/ParameterSet/interface/Registry.h"

struct ThreadWorker
{
  ThreadWorker(int n_to_add, int n_of_lookups, int ofs) :
    number_to_add(n_to_add),
    number_of_lookups(n_of_lookups),
    offset(ofs)
  { }

  void operator()();

  int number_to_add;
  int number_of_lookups;
  int offset;
};

void
ThreadWorker::operator()()
{
  edm::pset::Registry* reg = edm::pset::Registry::instance();
  assert (reg);

  // Add a bunch of items
  std::vector<edm::ParameterSetID> ids_issued;
  std::vector<edm::ParameterSet> psets;

  for (int i = offset; i < number_to_add+offset; ++i)
  {
      edm::ParameterSet ps;
      ps.addParameter<int>("i", i);
      ps.addUntrackedParameter<double>("d", 2.5);
      ids_issued.push_back(ps.id());
      psets.push_back(ps);
      //reg->insertMapped(ps);
  }
  edm::ParameterSet toplevel;
  toplevel.addParameter("guts", psets);
  edm::pset::loadAllNestedParameterSets(reg, toplevel);

  edm::ParameterSetID topid = edm::pset::getProcessParameterSetID(reg);
  assert( topid.isValid() );
  assert( topid == toplevel.id() );

  // Look up items we have just put in.
  typedef std::vector<edm::ParameterSetID>::const_iterator iter;
  for (iter i=ids_issued.begin(), e=ids_issued.end(); i!=e; ++i)
  {
      edm::ParameterSet ps = getParameterSet(*i);
      assert(ps.id() == *i);
      assert(ps.getParameter<int>("i") > 0);
      double mye = ps.getUntrackedParameter<double>("e", 1.0);
      assert ( std::abs(mye-1.0) < 1.0e-10 );
  }

  for (iter i=ids_issued.begin(), e=ids_issued.end(); i!=e; ++i)
  {
    edm::ParameterSet ps;
    assert(reg->getMapped(*i, ps));
    assert(ps.id() == *i);
  }

}

void work()
{
  edm::pset::Registry* reg = edm::pset::Registry::instance();
  assert (reg);

  // Launch a bunch of threads, which beat on the Registry...
  boost::thread_group threads;
  // don't make this too high, or resource exhaustion may kill
  // the whole test...
  const int NUM_THREADS = 10;
  const int NUM_PSETS_PER_THREAD = 500;
  const int NUM_LOOKUPS_PER_THREAD = 1000;
  for (int i = 0; i < NUM_THREADS; ++i)
    threads.create_thread(ThreadWorker(NUM_PSETS_PER_THREAD,
 				       NUM_LOOKUPS_PER_THREAD,
 				       NUM_PSETS_PER_THREAD));
  threads.join_all();
}

void work2()
{
  // This part is not multi-threaded; it checks only the return value
  // of the insertParameterSet member function.
  edm::pset::Registry* reg = edm::pset::Registry::instance();

  // Make a new ParameterSet, not like those already in the Registry.
  edm::ParameterSet ps;
  std::string value("and now for something completely different...");
  ps.addParameter<std::string>("s", value);

  // First call should insert the new ParameterSet; second call should
  // not.
  assert(reg->insertMapped(ps));
  assert(!reg->insertMapped(ps));

}


int main()
{
  int rc = 1;
  try
  {
      work();
      work2();

      // Look at what we have saved.
      //std::cout << "Here comes the registry..." << std::endl;
      //std::cout << *edm::pset::Registry::instance() << std::endl;
      //std::cout << "...done" << std::endl;
      rc = 0;
  }
  catch (edm::Exception& x)
  {
      std::cerr << "edm::Exception caught\n"
		<< x
		<< '\n';
      rc = -1;
  }
  catch (std::bad_alloc & x)
  {
      std::cerr << "bad alloc: " << x.what() << '\n';
      rc = -2;
  }

  catch (boost::thread_resource_error & x)
  {
      std::cerr << "resource error: " << x.what()
		<< "\nTry decreasing NUM_THREADS, and recompiling\n";	
      rc = -3;
  }


  catch (boost::lock_error & x)
  {
      std::cerr << "lock error: " << x.what() << '\n';
      rc = -4;
  }




  catch (...)
  {
      std::cerr << "Unknown exception caught\n";
      rc = 1;
  }
  return rc;
}
